#include <stdio.h>
#include "parti.h"

static void		_bsort_proc();
struct com_pattern	*sched_gather_proc();

SCHED *
sched_proc(proc,ndata,new_ndata,maxdim)
int	*proc, ndata, maxdim;
int	*new_ndata;		/* number of data after scatter_append */
{
	int	*pl,**perms,**buckets,*bsize;
	int	*GColor, i;
	struct	com_pattern *sr;
	SCHED	*sched;
 
#ifdef DEBUG 
	trace_enter("sched_proc");
#endif
	sched		= (SCHED *) malloc(sizeof(SCHED));
	pl		= (int *) malloc(sizeof(int)*ndata);
	perms		= (int **) malloc(sizeof(int *)*PARTI_numnodes);
	bsize		= (int *) malloc(sizeof(int)*PARTI_numnodes);

	sched->global_refs = NULL;
	sched->local_refs = NULL;
	sched->type	= NULL;
	_bsort_proc(PARTI_numnodes,ndata,proc,bsize,perms,pl);
	sr = sched_gather_proc(bsize,maxdim);

	GColor  = (int *) malloc(sizeof(int)*PARTI_numnodes);
	PairColor(GColor, PARTI_mynode);
	sr->gcolor    = GColor;

	sched->sr 	= sr;
	sched->perm 	= pl;
	sched->ndata 	= ndata;

	/* count the number of data */
	for (*new_ndata=0,i=0; i < PARTI_numnodes; i++)
		*new_ndata += sched->sr->send_n[i];

	free(perms);
#ifdef DEBUG
	trace_exit();
#endif
	return(sched);
}

/****************************************************************/
/* in:                                                          */
/*   nbuckets                                                   */
/*   ndata                                                      */
/*   key[ndata]                                                 */
/* out:                                                         */
/*   bsize[nbuckets]                                            */
/*   *perms[nbuckets]                                           */
/****************************************************************/
static void 
_bsort_proc(nbuckets,ndata,key,bsize,perms,plist)
int nbuckets,ndata,*key,*bsize,**perms,*plist;
{
	int	i, count, *bcount;

	/* find size of each bucket */
	for(i=0;i<nbuckets;i++) bsize[i] = 0;
	for(i=0;i<ndata;i++){
	  if(key[i]>=nbuckets){
	    fprintf(stderr,
	     "[%d]Error in bsort: not enough buckets [i-%d, key-%d, ndata-%d]\n", 
	       PARTI_mynode, i, key[i], ndata);
	    exit();
	  }
	  bsize[key[i]]++;
	}

	/* allocate memory for buckets */
	for(count=0,i=0; i<nbuckets ;i++){
		perms[i] = &plist[count];
		count += bsize[i];
	}

	/* group offsets of each element by destination procs */
	bcount = (int*)malloc(sizeof(int)*nbuckets);
	for(i=0;i<nbuckets;i++) bcount[i] = 0;
	for(i=0;i<ndata;i++) perms[key[i]][bcount[key[i]]++] = i;
	free(bcount);
}

#define	 SEND_MATR	510000	/* message id for matrix transpose */
#define	 SEND_SCHED	520000	/* message id for gathering schedules */

struct com_pattern *
sched_gather_proc(bsize,maxdim)
int	*bsize;
int	maxdim;
{
	int	i,j,k;
	int	n_msgs_to_send, n_msgs_to_recv;
	int	*send_n, **send_list;
	int	size, f_size, s_size;
	struct com_pattern *out_struct;

#ifdef DEBUG 
	trace_enter("sched_gather_proc"); 
#endif 

	/* find how many messages to be sent in exec phase */
	/* i.e. find the transpose of the original distributed matrix */
	send_n = (int *) malloc(sizeof(int)*PARTI_numnodes);
	for(i=1; i<PARTI_numnodes ;i++){
	  k = (PARTI_mynode+i) % PARTI_numnodes;
	  PARTI_csend(SEND_MATR+PARTI_mynode,&bsize[k],sizeof(int),k,PARTI_pid);
	}  
	/* receive msg sizes from other processors */
	for (i = 0; i < PARTI_numnodes; i++) {
	  if (i == PARTI_mynode) continue;
	  PARTI_crecv(SEND_MATR+i, &send_n[i], sizeof(int), i, PARTI_pid);
	}
	send_n[PARTI_mynode] = bsize[PARTI_mynode];

	/* put pointers into structre to be passed out */
	out_struct = (struct com_pattern*)malloc(sizeof(struct com_pattern));
	out_struct->buffer_size = 0;
	f_size = s_size =  0;
	for(i=0; i<PARTI_numnodes ;i++) {
		if (i == PARTI_mynode) continue;
		if(out_struct->buffer_size<send_n[i])
			out_struct->buffer_size = send_n[i];
		s_size += send_n[i];
		f_size += bsize[i];
	}
	size = (s_size < f_size)? f_size: s_size; 

	if (wbuffer->max_size < (size*maxdim*sizeof(double))) {
	  free(wbuffer->buffer); 
	  wbuffer->buffer = (char *)malloc((size*maxdim)*sizeof(double)); 
	  if (NULL == wbuffer->buffer)
	    fprintf(stderr,"ERROR:not enough memory -parti working buffer\n"),exit(0);
	  wbuffer->max_size = size * maxdim * sizeof(double);
	}
	n_msgs_to_recv = n_msgs_to_send = 0;
	for (i = 0; i < PARTI_numnodes; i++) {
		if (bsize[i]) n_msgs_to_send++;
		if (send_n[i]) n_msgs_to_recv++;
	}

	out_struct->n_msgs_to_recv = n_msgs_to_recv;
	out_struct->n_msgs_to_send = n_msgs_to_send;
	out_struct->send_n = send_n;
	out_struct->send_list = NULL;
	out_struct->fetch_n = bsize;
	out_struct->fetch_list = NULL;
	out_struct->dimen = maxdim;
	out_struct->send_size = NULL;
	out_struct->fetch_size = NULL;
	out_struct->total_send_size = s_size;
	out_struct->total_fetch_size = f_size;

#ifdef DEBUG
	trace_exit();
#endif
 
	return(out_struct);
}  

void free_lsched(sched)
SCHED *sched;
{
struct com_pattern *pattern;
int node;

pattern = sched->sr;

for(node=0; node<PARTI_numnodes; node++) {
  if (pattern->send_n[node] > 0 && pattern->send_list ) {
    if (pattern->send_list[node] != NULL) {
       free(pattern->send_list[node]);
       pattern->send_list[node] = NULL;
    }
  }
}

if ( pattern->fetch_list) free(pattern->fetch_list[0]);
free(pattern->send_n);
free(pattern->fetch_n);
if ( pattern->send_list)  free(pattern->send_list);
if (pattern->fetch_list)  free(pattern->fetch_list);
free(pattern->gcolor);
free(pattern);

if(sched->global_refs) free(sched->global_refs);
if(sched->local_refs) free(sched->local_refs);
free(sched->perm);
free(sched);
}

/****** WRAPPERS BEGINS ******************************************************/

SCHED*
sched_proc_noreuse_(proc,ndata,new_ndata,maxdim)
int	*proc,*ndata,*new_ndata,*maxdim;
{
	return(sched_proc(proc,*ndata,new_ndata,*maxdim));
}

int
sched_proc_(proc,ndata,new_ndata,maxdim)
int	*proc,*ndata,*new_ndata,*maxdim;
{
	SCHED	*sched;

	sched = sched_proc(proc,*ndata,new_ndata,*maxdim);
	return(register_schedule(sched));
}

void
free_sched_noreuse_(sched)
SCHED   **sched;
{
        free_lsched(*sched);
}

/****** WRAPPERS ENDS ********************************************************/
